# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: service.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)




DESCRIPTOR = _descriptor.FileDescriptor(
  name='service.proto',
  package='zmp',
  serialized_pb='\n\rservice.proto\x12\x03zmp\"C\n\x12ServiceRequestImpl\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x01(\x0c\x12\x0e\n\x06kwargs\x18\x03 \x01(\x0c\"f\n\x13ServiceResponseImpl\x12\x0f\n\x07service\x18\x01 \x01(\t\x12\x10\n\x08response\x18\x02 \x01(\x0c\x12,\n\texception\x18\x03 \x01(\x0b\x32\x19.zmp.ServiceExceptionImpl\"N\n\x14ServiceExceptionImpl\x12\x10\n\x08\x65xc_type\x18\x01 \x01(\t\x12\x11\n\texc_value\x18\x02 \x01(\x0c\x12\x11\n\ttraceback\x18\x03 \x01(\x0c\x32K\n\x0bServiceImpl\x12<\n\x07Service\x12\x17.zmp.ServiceRequestImpl\x1a\x18.zmp.ServiceResponseImpl')




_SERVICEREQUESTIMPL = _descriptor.Descriptor(
  name='ServiceRequestImpl',
  full_name='zmp.ServiceRequestImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='zmp.ServiceRequestImpl.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='args', full_name='zmp.ServiceRequestImpl.args', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='kwargs', full_name='zmp.ServiceRequestImpl.kwargs', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=22,
  serialized_end=89,
)


_SERVICERESPONSEIMPL = _descriptor.Descriptor(
  name='ServiceResponseImpl',
  full_name='zmp.ServiceResponseImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='service', full_name='zmp.ServiceResponseImpl.service', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='response', full_name='zmp.ServiceResponseImpl.response', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exception', full_name='zmp.ServiceResponseImpl.exception', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=91,
  serialized_end=193,
)


_SERVICEEXCEPTIONIMPL = _descriptor.Descriptor(
  name='ServiceExceptionImpl',
  full_name='zmp.ServiceExceptionImpl',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='exc_type', full_name='zmp.ServiceExceptionImpl.exc_type', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=unicode("", "utf-8"),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='exc_value', full_name='zmp.ServiceExceptionImpl.exc_value', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='zmp.ServiceExceptionImpl.traceback', index=2,
      number=3, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value="",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  serialized_start=195,
  serialized_end=273,
)

_SERVICERESPONSEIMPL.fields_by_name['exception'].message_type = _SERVICEEXCEPTIONIMPL
DESCRIPTOR.message_types_by_name['ServiceRequestImpl'] = _SERVICEREQUESTIMPL
DESCRIPTOR.message_types_by_name['ServiceResponseImpl'] = _SERVICERESPONSEIMPL
DESCRIPTOR.message_types_by_name['ServiceExceptionImpl'] = _SERVICEEXCEPTIONIMPL

class ServiceRequestImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEREQUESTIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceRequestImpl)

class ServiceResponseImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICERESPONSEIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceResponseImpl)

class ServiceExceptionImpl(_message.Message):
  __metaclass__ = _reflection.GeneratedProtocolMessageType
  DESCRIPTOR = _SERVICEEXCEPTIONIMPL

  # @@protoc_insertion_point(class_scope:zmp.ServiceExceptionImpl)


# @@protoc_insertion_point(module_scope)
